from os.path import join, dirname
import logging.config

import click
import yaml
import jsonschema

from .registrar import register_file, deregister_file, deregister_identifier
from .daemon import run_daemon
from .config import load_config


def setup_logging(debug=False):
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'brief': {
                'format': '%(levelname)s %(name)s: %(message)s'
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': 'DEBUG' if debug else 'INFO',
                'formatter': 'brief',
            }
        },
        'root': {
            'handlers': ['console'],
            'level': 'DEBUG' if debug else 'INFO',
        }
    })


def validate_config(config):
    with open(join(dirname(__file__), 'config-schema.yaml')) as f:
        schema = yaml.load(f)

    jsonschema.validate(config, schema)


@click.group()
def cli():
    pass


@cli.command(help='Run the registrar daemon, attaching to a Redis queue')
@click.option('--config-file', type=click.File('r'))
@click.option('--validate/--no-validate', default=False)
@click.option('--replace/--no-replace', default=False)
@click.option('--host', type=str)
@click.option('--port', type=int)
@click.option('--listen-queue', type=str)
@click.option('--deregister-queue', type=str)
@click.option('--progress-set', type=str)
@click.option('--failure-set', type=str)
@click.option('--success-set', type=str)
@click.option('--debug/--no-debug', default=False)
def daemon(config_file=None, validate=False, replace=False, host=None,
           port=None, listen_queue=None, progress_set=None,
           failure_set=None, success_set=None,
           deregister_queue=None, debug=False):
    setup_logging(debug)
    config = load_config(config_file)
    if validate:
        validate_config(config)
    run_daemon(
        config, replace, host, port, listen_queue, progress_set,
        failure_set, success_set, deregister_queue
    )


@cli.command(help='Run a single, one-off registration')
@click.argument('item', type=str)
@click.option('--config-file', type=click.File('r'))
@click.option('--validate/--no-validate', default=False)
@click.option('--replace/--no-replace', default=False)
@click.option('--debug/--no-debug', default=False)
def register(item, config_file=None, validate=False, replace=False,
             debug=False):
    setup_logging(debug)
    config = load_config(config_file)
    if validate:
        validate_config(config)

    register_file(config, item, replace)


@cli.command(help='Run a single, one-off de-registration')
@click.argument('--path', type=str)
@click.argument('--identifier', type=str)
@click.option('--config-file', type=click.File('r'))
@click.option('--validate/--no-validate', default=False)
@click.option('--debug/--no-debug', default=False)
def deregister(file_path=None, identifier=None, config_file=None,
               validate=False, debug=False):
    setup_logging(debug)
    config = load_config(config_file)
    if validate:
        validate_config(config)

    if file_path:
        deregister_file(config, file_path)
    elif identifier:
        deregister_identifier(config, identifier)


if __name__ == '__main__':
    cli()
