"""
config.py
============

Contains configuration functions for the registrar

"""

import os
from typing import TextIO
import re

import yaml


ENV_PATTERN = re.compile(r'.*?\${(\w+)}.*?')

def constructor_env_variables(loader: yaml.Loader, node: str):
    """Extracts the environment variable from the node's value

    Args:
        loader (yaml.Loader): the yaml loader
        node (str): the current node in the yaml

    Returns:
        str: the parsed string that contains the value of the environment
        variable
    """
    value = loader.construct_scalar(node)
    match = ENV_PATTERN.findall(value)  # to find all env variables in line
    if match:
        full_value = value
        for g in match:
            env_variable = os.environ.get(g, )
            if env_variable is not None:
                full_value = full_value.replace(
                    f'${{{g}}}', env_variable
                )
            else:
                return None
        return full_value
    return value


def load_config(input_file: TextIO) -> dict:
    """Attempts to load config from yaml file

    Args:
        input_file (TextIO): the file to be read

    Returns:
        dict: dictionary containing configuration values
    """
    tag = '!env'
    loader = yaml.SafeLoader

    # the tag will be used to mark where to start searching for the pattern
    # e.g. somekey: !env somestring${MYENVVAR}blah blah blah
    loader.add_implicit_resolver(tag, ENV_PATTERN, None)
    loader.add_constructor(tag, constructor_env_variables)

    return yaml.load(input_file, Loader=loader)
