"""
daemon.py
===========

Contains the daemon functions for the registrar

"""

import logging
import json

import redis

from .registrar import deregister_identifier, deregister_file, register_file


logger = logging.getLogger(__name__)


def run_daemon(config: dict, replace: bool, host: str, port: int, listen_queue: str, progress_set: str,
               failure_set: str, success_set: str, deregister_queue: str=None):
    """Run the registrar daemon, listening on a redis queue
        for files to be registered.

    Args:
        config (dict): configuration for the registrar
        replace (bool): replace the files
        host (str): redis host url
        port (int): redis port
        listen_queue (str): name of queue to listen for registration events
        progress_set (str): name of set for items in progress
        failure_set (str): name of set for failed items
        success_set (str): name of set for successfully processed items
        deregister_queue (str, optional): queue for deregistratiation of items. Defaults to None.

    Raises:
        Exception: Raised if path is already registered
    """
    # initialize the queue client
    client = redis.Redis(
        host=host, port=port, charset="utf-8", decode_responses=True
    )
    logger.debug("waiting for redis queue '%s'..." % listen_queue)
    queues = [listen_queue]
    if deregister_queue:
        queues.append(deregister_queue)

    while True:
        # fetch an item from the queue to be registered
        queue, value = client.brpop(queues)

        if queue == listen_queue:
            client.sadd(progress_set, value)
            # start the registration on that file
            try:
                register_file(config, value, replace)
                client.sadd(success_set, value)
                client.srem(progress_set, value)
            except Exception as e:
                if 'is already registered' not in "%s" % e:
                    # do not add to failure if skipped due to already
                    # registered
                    client.sadd(failure_set, value)
                client.srem(progress_set, value)
                logger.exception(e)
        elif queue == deregister_queue:
            parsed = json.loads(value)
            try:
                if 'identifier' in parsed:
                    deregister_identifier(config, parsed['identifier'])
                elif 'url' in parsed:
                    deregister_file(config, parsed['url'])
                else:
                    raise Exception(
                        "Neither 'identifer' nor 'url' passed to "
                        "deregistration"
                    )
            except Exception as e:
                logger.exception(e)
