"""
registrar.py
=============

Contains all functions relevant for registration

"""
import logging
from typing import List
import json

from .source import get_source
from .backend import get_backends
from .exceptions import RegistrationError
from .utils import import_by_path
from pystac import Item

logger = logging.getLogger(__name__)


def register_file(config: dict, item: str, replace: bool = False):
    """Handle registration of a stac item

    Args:
        config (dict): configuration of the registrar
        item (str): STAC item to register
        replace (bool, optional): replace the file in the endpoint or not.
                                  Defaults to False.

    Raises:
        RegistrationError: Raised if file is already registered and replace is
                           not turned on

    Returns:
        list: List of contexts
    """
    item = Item.from_dict(json.loads(item))
    source = get_source(config, item)
    logger.info(f"Handling '{item!r}'.")

    for pre_handler in get_pre_handlers(config):
        pre_handler(config, item)

    try:
        for backend in get_backends(config):
            if backend.exists(source, item):
                if replace:
                    logger.info(f"Replacing '{item!r}'.")
                    backend.register(source, item, replace=True)
                else:
                    raise RegistrationError(
                        f'Object {item!r} is already registered'
                    )
            else:
                logger.info(f"Registering '{item!r}'.")
                backend.register(source, item, replace=False)
    except Exception as e:
        for error_handler in get_error_handlers(config):
            error_handler(config, item, e)
        raise
    else:
        for post_handler in get_post_handlers(config):
            post_handler(config, item)

        logger.info(
            f"Successfully {'replaced' if replace else 'registered'} "
            f"'{item!r}'"
        )

    return item


def deregister_identifier(config, identifier: str):
    logger.info(f"Handling deregistration of identifier '{identifier}'.")
    contexts = [Context(
        identifier=identifier,
        path=None,
        scheme=None
    )]
    return _deregister_contexts(config, contexts)


def deregister_file(config: dict, path: str):
    logger.info(f"Handling deregistration of path '{path}'.")
    source = get_source(config, path)
    scheme = get_scheme(config, path)
    contexts = scheme.get_context(source, path)
    return _deregister_contexts(config, contexts)


def _deregister_contexts(config, items: List[Item]):
    if isinstance(items, Item):
        items = [items]

    for item in items:
        for pre_handler in get_deregister_pre_handlers(config):
            pre_handler(config, context, context)

        try:
            for backend in get_backends(config, context):
                backend.deregister(context)
        except Exception as e:
            for error_handler in get_deregister_error_handlers(config):
                error_handler(config, context, context, e)
            raise
        else:
            for post_handler in get_deregister_post_handlers(config):
                post_handler(config, context, context)

            logger.info(f"Successfully deregistered '{context.identifier}'")

    return items


def _get_handlers(config, name):
    handlers = []
    for handler_def in config.get(name, []):
        handler_cls = import_by_path(handler_def['path'])
        handlers.append(
            handler_cls(
                *handler_def.get('args', []),
                **handler_def.get('kwargs', []),
            )
        )

    return handlers


def get_pre_handlers(config):
    return _get_handlers(config, 'pre_handlers')


def get_post_handlers(config):
    return _get_handlers(config, 'post_handlers')


def get_error_handlers(config):
    return _get_handlers(config, 'error_handlers')


def get_deregister_pre_handlers(config):
    return _get_handlers(config, 'deregister_pre_handlers')


def get_deregister_post_handlers(config):
    return _get_handlers(config, 'deregister_post_handlers')


def get_deregister_error_handlers(config):
    return _get_handlers(config, 'deregister_error_handlers')
